/*QuadPlayer Song Optimizer
Copyright (c) 2005 James Byerly

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
#include <cstdio>
#include <set>
#include <limits>
#include "QPSong.h"
#include "PatternInfo.h"
using namespace std;

//These types are used to speed up searching for pattersn by storing where we have searched already;
typedef vector<bool> SectionPatternsLocationVec;
typedef vector<SectionPatternsLocationVec> SongPatternLocationsVec;

//Initialize the SongPatternLocationsVec to the same dimensions as SectionsNotes, with value = false
void InitializeSongPatternLocations(const QPSong &p_TheSong, SongPatternLocationsVec &p_TheSongPatternLocations)
{
	const size_t NumSections = p_TheSong.SectionsNotes.size();
	p_TheSongPatternLocations.clear();
	p_TheSongPatternLocations.resize(NumSections, SectionPatternsLocationVec(0));
	for(size_t WhichSection = 0; WhichSection < NumSections; ++WhichSection)
	{
		const size_t NumNotesInSection = p_TheSong.SectionsNotes[WhichSection].size();
		p_TheSongPatternLocations[WhichSection].clear();
		p_TheSongPatternLocations[WhichSection].resize(NumNotesInSection, false);
	}
}

//Find all occurances of ThePattern in TheSong,
//using TheSongPatternLocations to quick-skip already searched locations and to avoid adding a spot twice
void FindPatternInstances(const QPSong &p_TheSong, PatternInfo &p_ThePattern, SongPatternLocationsVec &p_TheSongPatternLocations)
{
	if(p_TheSongPatternLocations[p_ThePattern.Start.Section][p_ThePattern.Start.Note] != false)
	{
		//Already found a starting where ThePattern starts, so don't bother looking for the pattern again
		return;
	}
	if(p_ThePattern.Start.Note == 3704)
	{
		_asm int 3;
	}
	const QPSong::SectionVec &PatternSectionsNotes = p_TheSong.SectionsNotes[p_ThePattern.Start.Section];
	for(size_t LookInSection = 0;
		LookInSection < p_TheSong.SectionsNotes.size();
		++LookInSection)
	{
		const QPSong::SectionVec &CurrentSectionsNotes = p_TheSong.SectionsNotes[LookInSection];
		const size_t CurrentSectionSize = CurrentSectionsNotes.size();
		if(p_ThePattern.Length > CurrentSectionSize)
		{
			continue;
		}
		//Loop through section counting occurances
		for(size_t FindOccuranceAt = 0;
			FindOccuranceAt < (CurrentSectionSize - p_ThePattern.Length);
			/*Increment depends on whether match found or not*/)
		{
			if(p_TheSongPatternLocations[LookInSection][FindOccuranceAt] != false)
			{
				//Already found a pattern here of this length, so skip it
				FindOccuranceAt += p_ThePattern.Length;
				continue;
			}
			bool bFoundMatch = true;
			//Loop through the notes in the pattern and compare them to FindOccuranceAt
			for(size_t WhichNote = 0; WhichNote < p_ThePattern.Length; ++WhichNote)
			{
				const QPNote PatternNote = PatternSectionsNotes[p_ThePattern.Start.Note + WhichNote];
				const QPNote CurrentNote = CurrentSectionsNotes[FindOccuranceAt + WhichNote];
				if(PatternNote != CurrentNote)
				{
					bFoundMatch = false;
					break;
				}
			}
			if(bFoundMatch)
			{
				//We just found a match, so add to occurance locations and say we found something here
				p_TheSongPatternLocations[LookInSection][FindOccuranceAt] = true;
				p_ThePattern.OccuranceLocation.push_back(QPNoteCoord(LookInSection, FindOccuranceAt));
				FindOccuranceAt += p_ThePattern.Length;
			}
			else
			{
				++FindOccuranceAt;
			}
		}
	}
}

//Find occurances of ThePattern in TheSong starting at PossibleLocations,
//using TheSongPatternLocations to quick-skip already searched locations and to avoid adding a spot twice
//Only test last note because we're guaranteed that all matches prior to that
void ExpandPatternInstance(const QPSong &p_TheSong,
						   const PatternInfo::OccuranceLocationVec &p_PossibleLocations,
						   PatternInfo &p_ThePattern,
						   SongPatternLocationsVec &p_TheSongPatternLocations)
{
	if(p_TheSongPatternLocations[p_ThePattern.Start.Section][p_ThePattern.Start.Note] != false)
	{
		//Already found a starting where ThePattern starts, so don't bother looking for the pattern again
		return;
	}
	typedef PatternInfo::OccuranceLocationVec::const_iterator LocationConstIterator;
	LocationConstIterator StartLocation = p_PossibleLocations.begin();
	const LocationConstIterator LastStartLocation = p_PossibleLocations.end();

	QPNoteCoord PreviousLocation = p_ThePattern.Start;

	const QPNote &PatternEndNote = p_TheSong.SectionsNotes[p_ThePattern.Start.Section][p_ThePattern.Start.Note + p_ThePattern.Length - 1];
	for(;StartLocation != LastStartLocation; ++StartLocation)
	{
		const QPNoteCoord &CurrentLocation = *StartLocation;
		if(p_TheSongPatternLocations[CurrentLocation.Section][CurrentLocation.Note] != false)
		{
			//Already found a pattern match here, so skip this location (two patterns can't occupy same space exactly)
			continue;
		}
		//Make sure the section of CurrentLocation is long enough to hold the end note position
		if(p_TheSong.SectionsNotes[CurrentLocation.Section].size()
			<= (CurrentLocation.Note + p_ThePattern.Length))
		{
			continue;
		}
		//Make sure this pattern doesn't overlap with the last found occurance (it can, now that length is +1)
		//Note that this is a shortcut that works only because of the SongPatternLocation vector
		//Without it, it would
		if((PreviousLocation.Section == CurrentLocation.Section)
			&& ((PreviousLocation.Note + p_ThePattern.Length) > CurrentLocation.Note)
			&& (PreviousLocation.Note < CurrentLocation.Note))
		{
			PreviousLocation = CurrentLocation;
			continue;
		}
		PreviousLocation = CurrentLocation;
		//Compare the notes to see if the pattern is here 
		const QPNote &CurrentEndNote = p_TheSong.SectionsNotes[CurrentLocation.Section][CurrentLocation.Note + p_ThePattern.Length - 1];
		if(PatternEndNote == CurrentEndNote)
		{
			p_ThePattern.OccuranceLocation.push_back(CurrentLocation);
			p_TheSongPatternLocations[CurrentLocation.Section][CurrentLocation.Note] = true;
		}
	}
}

void ApplyPatternToSong(const PatternInfo &p_BestPattern, QPSong &p_TheSong)
{
	//Store the new sections and notes temporarily in these local variables
	QPSong::SectionOrderVec NewSectionOrder;
	QPSong::SectionsNotesVec NewSectionsNotes;

	//Create a vector that holds how to change the Section Order
	//since a section can be cut into many depend on where patterns occur
	typedef vector<size_t> SectionReplacementVec;
	typedef vector<SectionReplacementVec> SongSectionReplacementVec;
	SongSectionReplacementVec SongSectionReplacement;

	//Initialize some variables
	const size_t NumOriginalSections = p_TheSong.SectionsNotes.size();
	SongSectionReplacement.resize(NumOriginalSections);

	//Copy the pattern to the first section (number 0)
	NewSectionsNotes.push_back(QPSong::SectionVec());//Insert blank note list
	for(size_t PatternNote = 0; PatternNote < p_BestPattern.Length; ++PatternNote)
	{
		//Add a single note
		const QPNote &TheNote = p_TheSong.SectionsNotes[p_BestPattern.Start.Section][p_BestPattern.Start.Note + PatternNote];
		NewSectionsNotes[0].push_back(TheNote);
	}

	//Initialize iterators to move through the pattern occurance locations
	typedef PatternInfo::OccuranceLocationVec::const_iterator LocationConstIterator;
	LocationConstIterator PatternLocation = p_BestPattern.OccuranceLocation.begin();
	const LocationConstIterator LastPatternLocation = p_BestPattern.OccuranceLocation.end();

	//Go through each original section and (if needed) break it into multiple sections around the pattern
	for(size_t CurrentOriginalSection = 0; CurrentOriginalSection < NumOriginalSections; ++CurrentOriginalSection)
	{
		SongSectionReplacement[CurrentOriginalSection].clear();
		if((PatternLocation == LastPatternLocation)
			|| (PatternLocation->Section != CurrentOriginalSection))
		{
			//This section doesn't have the pattern in it, so it's replacement is just an adjusted section number
			SongSectionReplacement[CurrentOriginalSection].push_back(NewSectionsNotes.size());
			NewSectionsNotes.push_back(p_TheSong.SectionsNotes[CurrentOriginalSection]);
		}
		else
		{
			size_t CurrentOriginalNote = 0;
			//Keep processing until the next PatternLocation is in a different section
			do{
				if(CurrentOriginalNote != PatternLocation->Note)
				{
					const size_t CurrentNewSection = NewSectionsNotes.size();
					NewSectionsNotes.push_back(QPSong::SectionVec());//Insert blank note list
					for(;CurrentOriginalNote < PatternLocation->Note; ++CurrentOriginalNote)
					{
						NewSectionsNotes[CurrentNewSection].push_back(p_TheSong.SectionsNotes[CurrentOriginalSection][CurrentOriginalNote]);
					}
					SongSectionReplacement[CurrentOriginalSection].push_back(CurrentNewSection);
				}
				SongSectionReplacement[CurrentOriginalSection].push_back(0);
				CurrentOriginalNote += p_BestPattern.Length;
				++PatternLocation;
			}while((PatternLocation != LastPatternLocation)
				&& (PatternLocation->Section == CurrentOriginalSection));
			//If the last pattern occurance wasn't at the end of the section, make sure the end of it is copied
			if(CurrentOriginalNote < p_TheSong.SectionsNotes[CurrentOriginalSection].size())
			{
				const size_t LastNote = p_TheSong.SectionsNotes[CurrentOriginalSection].size();
				const size_t CurrentNewSection = NewSectionsNotes.size();
				NewSectionsNotes.push_back(QPSong::SectionVec());//Insert blank note list
				for(;CurrentOriginalNote < LastNote; ++CurrentOriginalNote)
				{
					NewSectionsNotes[CurrentNewSection].push_back(p_TheSong.SectionsNotes[CurrentOriginalSection][CurrentOriginalNote]);
				}
				SongSectionReplacement[CurrentOriginalSection].push_back(CurrentNewSection);
			}
		}
	}
	//Use the SongSectionReplacement to fix up the SectionOrder vector
	for(size_t CurrentSectionOrderIndex = 0;
		CurrentSectionOrderIndex < p_TheSong.SectionOrder.size();
		++CurrentSectionOrderIndex)
	{
		const size_t SectionToReplace = p_TheSong.SectionOrder[CurrentSectionOrderIndex];
		const SectionReplacementVec &SectionReplacement = SongSectionReplacement[SectionToReplace];
		for(size_t SectionOrderReplacement = 0;
			SectionOrderReplacement < SectionReplacement.size();
			++SectionOrderReplacement)
		{
			NewSectionOrder.push_back(SongSectionReplacement[SectionToReplace][SectionOrderReplacement]);
		}
	}
	p_TheSong.SectionOrder = NewSectionOrder;
	p_TheSong.SectionsNotes = NewSectionsNotes;
}

bool OptimizeSongFile(const char *p_InFileName, const char *p_OutFileName)
{
	QPSong TheSong;
	if(!TheSong.ReadFromFile(p_InFileName))
	{
		return false;
	}
	TheSong.Expand();
	QPSong DebugSongBackup = TheSong;

	//Easy way to find patterns - pick a 2-note sequence
	// go through the whole song to see how many times it appears
	//All 3-note patterns must start with a 2-note pattern, so use that to speed up later passes
	//Then use the best pattern to make a new section. Repeat until no patterns save bytes
	while(true)//Loop until mid-loop break conditions
	{
		//These sets will be sorted according to quality of the pattern, which is a rough estimate
		typedef vector<PatternInfo> SongPatternsVec;
		SongPatternsVec ShortSongPatterns, LongSongPatterns;
		//This pattern info stores the best pattern known so far
		PatternInfo BestPattern;
		//This vector is used to track where patterns have been found already (so the same place isn't searched twice)
		SongPatternLocationsVec SongPatternLocations;

		BestPattern.Length = 0;
		BestPattern.BytesSaved = numeric_limits<int>::min();

		InitializeSongPatternLocations(TheSong, SongPatternLocations);

		//First find all patterns of length 2
		//Loop through all the sections in the song
		for(size_t PatternSection = 0;
			PatternSection < TheSong.SectionsNotes.size();
			++PatternSection)
		{
			const size_t SectionSize = TheSong.SectionsNotes[PatternSection].size();
			const size_t PatternLength = 2;
			//If the pattern couldn't fit in the section, skip this iteration
			if(SectionSize < PatternLength)
			{
				continue;
			}
			//Loop through all possible starting position in PatternSection
			for(size_t PatternNote = 0;
				PatternNote < (SectionSize - PatternLength);
				++PatternNote)
			{
				PatternInfo CurrentPattern;
				CurrentPattern.Length = PatternLength;
				CurrentPattern.Start.Section = PatternSection;
				CurrentPattern.Start.Note = PatternNote;
				//Because we search every section for the pattern,
				// it will be found at it's originating point also
				CurrentPattern.OccuranceLocation.clear();
				
				//Find all instances of this pattern in the song
				FindPatternInstances(TheSong, CurrentPattern, SongPatternLocations);

				//If this pattern occurs more than once, save it to search for longer versions later
				if(CurrentPattern.OccuranceLocation.size() > 1)
				{
					CurrentPattern.CalcBytesSaved(TheSong);
					if(CurrentPattern > BestPattern)
					{
						BestPattern = CurrentPattern;
					}
					ShortSongPatterns.push_back(CurrentPattern);
				}
			}
		}
		//If no patterns are found, no point in searching for longer patterns
		if(ShortSongPatterns.size() == 0)
		{
			break;
		}
		//Loop through all found patterns and see if they make longer patterns also
		do{
			InitializeSongPatternLocations(TheSong, SongPatternLocations);
			//For each pattern in ShortSongPatterns
			SongPatternsVec::iterator ExtendPattern = ShortSongPatterns.begin();
			const SongPatternsVec::iterator LastExtendPattern = ShortSongPatterns.end();
			for(;ExtendPattern != LastExtendPattern; ++ExtendPattern)
			{
				//For each location in the current pattern
				typedef PatternInfo::OccuranceLocationVec::iterator LocationIterator;
				LocationIterator StartLocation = ExtendPattern->OccuranceLocation.begin();
				const LocationIterator LastStartLocation = ExtendPattern->OccuranceLocation.end();
				for(;StartLocation != LastStartLocation; ++StartLocation)
				{
					PatternInfo CurrentPattern;
					CurrentPattern = *ExtendPattern;
					CurrentPattern.Start = *StartLocation;
					CurrentPattern.Length++;
					CurrentPattern.OccuranceLocation.clear();

					//Make sure this section is long enough to have this pattern
					if(TheSong.SectionsNotes[CurrentPattern.Start.Section].size()
						<= (CurrentPattern.Start.Note + CurrentPattern.Length))
					{
						//If not, try next start location
						continue;
					}

					ExpandPatternInstance(TheSong, ExtendPattern->OccuranceLocation, CurrentPattern, SongPatternLocations);

					//If this pattern occurs more than once, save it to search for longer versions later
					if(CurrentPattern.OccuranceLocation.size() > 1)
					{
						CurrentPattern.CalcBytesSaved(TheSong);
						if(CurrentPattern > BestPattern)
						{
							BestPattern = CurrentPattern;
						}
						LongSongPatterns.push_back(CurrentPattern);
					}
				}
			}
			ShortSongPatterns = LongSongPatterns;
			LongSongPatterns.clear();
		}while(ShortSongPatterns.size() > 0);

		//If no patterns save bytes, don't bother inserting patterns into the song
		if(BestPattern.BytesSaved <= 0)
		{
			break;
		}
		ApplyPatternToSong(BestPattern, TheSong);
	}
	/*
	TheSong.Expand();
	if(TheSong.SectionsNotes[0].size() != DebugSongBackup.SectionsNotes[0].size())
	{
		cout << "Error" << endl;
	}
	else
	{
		const QPSong::SectionVec &DebugBackupNotes = DebugSongBackup.SectionsNotes[0];
		const QPSong::SectionVec &RealSongNotes = TheSong.SectionsNotes[0];
		for(size_t WhichNote = 0; WhichNote < RealSongNotes.size(); ++WhichNote)
		{
			if(DebugBackupNotes[WhichNote] != RealSongNotes[WhichNote])
			{
				cout << WhichNote;
				cout << " D(" << int(DebugBackupNotes[WhichNote].Left1);
				cout << "," << int(DebugBackupNotes[WhichNote].Left2);
				cout << "," << int(DebugBackupNotes[WhichNote].Right1);
				cout << "," << int(DebugBackupNotes[WhichNote].Right2);
				cout << "," << int(DebugBackupNotes[WhichNote].Duration) << ")";
				cout << " R(" << int(RealSongNotes[WhichNote].Left1);
				cout << "," << int(RealSongNotes[WhichNote].Left2);
				cout << "," << int(RealSongNotes[WhichNote].Right1);
				cout << "," << int(RealSongNotes[WhichNote].Right2);
				cout << "," << int(RealSongNotes[WhichNote].Duration) << ")" << endl;
			}
		}
	}
	*/
	return TheSong.WriteToFile(p_OutFileName);
}

int main(int argc, char *argv[])
{
	if(argc < 3)
	{
		printf("QTOpt InFile OutFile\n");
	}
	else
	{
		return !OptimizeSongFile(argv[1], argv[2]);
	}
}