/*QuadPlayer Song Optimizer
Copyright (c) 2005 James Byerly

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
#include <vector>
#include <map>
#include <set>
#include <string>
#include "QPSong.h"
using namespace std;

namespace /*For this file only*/
{
	template <typename VecType>
	string ReadString(const VecType &p_Vec, typename VecType::size_type p_StartPosition)
	{
		char ChString[2] = {0, 0};
		string ReturnValue;

		while((p_StartPosition != p_Vec.size())
			&& (p_Vec[p_StartPosition] != 0))
		{
			ChString[0] = p_Vec[p_StartPosition];
			++p_StartPosition;
			ReturnValue += string(ChString);
		}
		return ReturnValue;
	}
}

bool QPSong::ReadFromFile(const char *p_FileName)
{
	if(!ProgramFile.ReadProgramFile(p_FileName))
	{
		//Open failed
		return false;
	}
	TI83PlusProgram::ProgramDataVec &ProgramData = ProgramFile.ProgramData;

	size_t CurrentPosition = 0;
	
	//@@Read Strings
	//Read Signature ("QUAD1")
	Signature = ReadString(ProgramData, CurrentPosition);
	if(Signature != "QUAD1")
	{
		return false;
	}
	CurrentPosition += Signature.size() + 1; //+1 for terminating NUL character
	//Read Title
	Title = ReadString(ProgramData, CurrentPosition);
	CurrentPosition += Title.size() + 1; //+1 for terminating NUL character
	//Read Artist
	Artist = ReadString(ProgramData, CurrentPosition);
	CurrentPosition += Artist.size() + 1; //+1 for terminating NUL character
	//Read Album
	Album = ReadString(ProgramData, CurrentPosition);
	CurrentPosition += Album.size() + 1; //+1 for terminating NUL character

	//@@Read Sections
	//First load the addresses into the list
	int AddressIndex = 0;
	typedef map<int,int> SectionIndexMap;
	SectionIndexMap SectionIndex;//maps addresses to integer index (0=first)
	int Address;

	do{
		if(CurrentPosition + 1 > ProgramData.size())
		{
			//Serious error - should never occur
			return false;
		}
		//Convert address
		Address = ProgramData[CurrentPosition] + 
			(int(ProgramData[CurrentPosition + 1]) << 8);
		CurrentPosition += 2;

		if(Address == 0)
		{
			break;
		}
		SectionOrder.push_back(Address);

		SectionIndexMap::const_iterator AlreadyIndexed = SectionIndex.find(Address);
		//If not already indexed, add to index map
		if(AlreadyIndexed == SectionIndex.end())
		{
			//Sections are numbered in order they appear - first is 0, etc
			SectionIndex[Address] = AddressIndex;
			++AddressIndex;
		}
	}while(Address != 0);//Will always break earlier in loop
	//Zero SectionsNotes so we can tell if we loaded a Section's notes already
	SectionsNotes.clear();
	SectionsNotes.resize(SectionIndex.size());
	//Load notes, then convert address to index so that SectionOrder contains indexes
	for(size_t WhichSection = 0; WhichSection < SectionOrder.size(); ++WhichSection)
	{
		//Convert Address -> Index
		Address = SectionOrder[WhichSection];
		SectionIndexMap::const_iterator Index = SectionIndex.find(Address);
		if(Index == SectionIndex.end())
		{
			//Serious error - section not found in map (should never occur)
			return false;
		}
		int CurrentSectionIndex = Index->second;
		SectionOrder[WhichSection] = CurrentSectionIndex;
		//Check if we already loaded this section's notes:			
		vector<QPNote> &CurrentSectionsNotes = SectionsNotes[CurrentSectionIndex];
		//If not, load them now
		if(CurrentSectionsNotes.size() == 0)
		{
			CurrentPosition = Address;
			const QPNote ZeroNote = {0,0,0,0,0};
			QPNote CurrentNote;
			//Load notes
			do{
				//Order L1 L2 R1 R2
				CurrentNote.Left1 = ProgramData[CurrentPosition++];
				CurrentNote.Left2 = ProgramData[CurrentPosition++];
				if(CurrentNote.Left1 < CurrentNote.Left2)
				{
					swap(CurrentNote.Left1, CurrentNote.Left2);
				}
				CurrentNote.Right1 = ProgramData[CurrentPosition++];
				CurrentNote.Right2 = ProgramData[CurrentPosition++];
				if(CurrentNote.Right1 < CurrentNote.Right2)
				{
					swap(CurrentNote.Right1, CurrentNote.Right2);
				}
				CurrentNote.Duration = ProgramData[CurrentPosition]
					+ (WORD(ProgramData[CurrentPosition + 1]) << 8);
				CurrentPosition += 2;
				CurrentSectionsNotes.push_back(CurrentNote);
			}while((CurrentNote != ZeroNote) && (CurrentPosition != ProgramData.size()));
			//Remove ending 0s
			CurrentSectionsNotes.pop_back();
		}
	}
	//Clear program data - will be reconstructed when saving
	ProgramData.clear();
	return true;
}

bool QPSong::WriteToFile(const char *p_FileName)
{
	TI83PlusProgram::ProgramDataVec &ProgramData = ProgramFile.ProgramData;
	ProgramData.clear();
	for(size_t SignatureIndex = 0; SignatureIndex < Signature.size(); ++SignatureIndex)
	{
		ProgramData.push_back(Signature[SignatureIndex]);
	}
	ProgramData.push_back(0);
	for(size_t TitleIndex = 0; TitleIndex < Title.size(); ++TitleIndex)
	{
		ProgramData.push_back(Title[TitleIndex]);
	}
	ProgramData.push_back(0);
	for(size_t ArtistIndex = 0; ArtistIndex < Artist.size(); ++ArtistIndex)
	{
		ProgramData.push_back(Artist[ArtistIndex]);
	}
	ProgramData.push_back(0);
	for(size_t AlbumIndex = 0; AlbumIndex < Album.size(); ++AlbumIndex)
	{
		ProgramData.push_back(Album[AlbumIndex]);
	}
	ProgramData.push_back(0);
	//Calculate Section Addresses:
	vector<WORD> SectionAddresses;
	WORD CurrentSectionAddress = WORD(SectionOrder.size() + 1) * 2 +
		//Take into account size of strings
		WORD(Signature.size() + Title.size() + Artist.size() + Album.size() + 4);
	SectionAddresses.resize(SectionsNotes.size());
	for(size_t WhichSection = 0; WhichSection < SectionsNotes.size(); ++WhichSection)
	{
		SectionAddresses[WhichSection] = CurrentSectionAddress;
		CurrentSectionAddress += WORD(SectionsNotes[WhichSection].size() + 1) * 6;
	}
	//Output Section Addresses
	for(size_t SectionOrderIndex = 0; SectionOrderIndex < SectionOrder.size(); ++SectionOrderIndex)
	{
		const WORD SectionAddress = WORD(SectionAddresses[SectionOrder[SectionOrderIndex]]);
		ProgramData.push_back(BYTE(SectionAddress & 0xFF));
		ProgramData.push_back(BYTE(SectionAddress >> 8));
	}
	ProgramData.push_back(0);
	ProgramData.push_back(0);
	//Output Section Notes
	for(size_t SectionsNotesIndex = 0; SectionsNotesIndex < SectionsNotes.size(); ++SectionsNotesIndex)
	{
		for(size_t NotesIndex = 0; NotesIndex < SectionsNotes[SectionsNotesIndex].size(); ++NotesIndex)
		{
			const QPNote &CurrentNote = SectionsNotes[SectionsNotesIndex][NotesIndex];
			ProgramData.push_back(CurrentNote.Left1);
			ProgramData.push_back(CurrentNote.Left2);
			ProgramData.push_back(CurrentNote.Right1);
			ProgramData.push_back(CurrentNote.Right2);
			ProgramData.push_back(BYTE(CurrentNote.Duration & 0xFF));
			ProgramData.push_back(BYTE(CurrentNote.Duration >> 8));
		}
		ProgramData.push_back(0);
		ProgramData.push_back(0);
		ProgramData.push_back(0);
		ProgramData.push_back(0);
		ProgramData.push_back(0);
		ProgramData.push_back(0);
	}
	return ProgramFile.WriteProgramFile(p_FileName);
}

void QPSong::Expand(void)
{
	//Convert PlayList/Sections structure into simple vector of notes in a single section
	vector<QPNote> WholeSong;

	for(size_t PlayingPart = 0;
		PlayingPart < SectionOrder.size();
		++PlayingPart)
	{
		int PlayingSection = SectionOrder[PlayingPart];
		for(size_t Notes = 0;
			Notes < SectionsNotes[PlayingSection].size();
			++Notes)
		{
			QPNote &TheNote = SectionsNotes[PlayingSection][Notes];
			WholeSong.push_back(TheNote);
		}
	}
	//Only section is #0
	SectionOrder.clear();
	SectionsNotes.clear();

	SectionOrder.push_back(0);
	SectionsNotes.push_back(WholeSong);
}